﻿namespace HeuristicLab.Algorithms.PushGP.Expressions {
  using HeuristicLab.Algorithms.PushGP.Generators;
  using HeuristicLab.Algorithms.PushGP.Interpreter;
  using HeuristicLab.Algorithms.PushGP.Stack;

  /// <summary>
  /// ERC expressions are only used for code generation.
  /// </summary>
  /// <typeparam name="T"></typeparam>
  public abstract class ErcExpression<T> : StatefullExpression<T> {
    protected ErcExpression(T value, string stringRepresentation) : base(value, stringRepresentation) { }

    protected void Eval(IStack<T> stack) {
      stack.Push(this.State);
    }
  }

  public class IntegerErcExpression : ErcExpression<long> {
    public IntegerErcExpression() : base(IntegerGenerator.RandomInteger(), "INTEGER.ERC") { }

    public override void Eval(IPushGpInterpreter interpreter) {
      this.Eval(interpreter.IntegerStack);
    }
  }

  public class FloatErcExpression : ErcExpression<double> {
    public FloatErcExpression() : base(FloatGenerator.RandomFloat(), "FLOAT.ERC") { }

    public override void Eval(IPushGpInterpreter interpreter) {
      this.Eval(interpreter.FloatStack);
    }
  }

  public class BooleanErcExpression : ErcExpression<bool> {
    public BooleanErcExpression() : base(BooleanGenerator.RandomBoolean(), "BOOLEAN.ERC") { }

    public override void Eval(IPushGpInterpreter interpreter) {
      this.Eval(interpreter.BooleanStack);
    }
  }

  public class NameErcExpression : ErcExpression<string> {
    public NameErcExpression() : base(NameGenerator.RandomName(), "NAME.ERC") { }

    public override void Eval(IPushGpInterpreter interpreter) {
      this.Eval(interpreter.NameStack);
    }
  }
}